import es, urllib, hashlib, os.path, sys, time, gamethread
from threading import Thread
import traceback as tb_module

###                                 ~README~
###                Errorlib for EventScripts 2 (RC 3) by Ojii
###
### Features :
###     - Sends emails to the scriptauthor
###     - Saves a errorlog on the server
###     - Only sends an error once
###     - Does not require an own SMTP
###
### Usage :
###   on load():
###     import errorlib
###     errorlib.registerAddon(info,email,language)
###   on unload(): (optional but recommended)
###     errorlib.unregisterAddon(basename)
###
### Thanks to :
###     Saul`for the help with the PHP backend
###     Everyone else on #eventscripts that helped
###     Mattie for obvious reasons
###     SuperDave for pointing at the weaknesses of my script which allowed me to improve it
###
### Changelog :
###   RC 4:
###     - Fixed unregisterAddon
###   RC 3:
###     - Changed scanError to return a list of involved registered addons.
###     - Changed unregisterAddon to support info instead of basename (unifies registring and unregistring).
###     - Changed registerAddon to support dict input instead of es.AddonInfo.
###     - Changed autoPrune to be done gamethread delayed every hour.
###     - autoPrune now also removes unloaded addons from the reggedAddons.
###     - Unified-Traceback now hashes srcds side instead of php side when longer than 32 Bytes to reduce traffic.
###     - Added configuration variable "filetimeout" which sets the autoPrune limit.
###     - Fixed tracebackUnify and loadedAddonsList to work with Linux.
###     - Minor changes: PHP Backend, Layout for logs and emails, default method is now email.
###   RC 2:
###     - Auto prune.
###     - improved anti-spam (gamepath and error description free identification of errors).
###     - unregisterAddon (to force autoprune).
###     - code improvements (suggested by SuperDave).
###   RC 1:
###     - Some minor changes to the code (especially addonslist).
###     - Supports german emails now.
###     - PHP side spam protection.
###     - Debug messages are now debug level 2 and above.
###     - Threading.
###                                 ~README~

reggedAddons = {}
gt = False
gamedir = str(es.ServerVar('eventscripts_gamedir')).replace('\\','/')

### This code from es.py, it handles errors... ###
def except_hook(type, value, ttemplateback):
	error_list = tb_module.format_exception(type, value, ttemplateback)
	for message in error_list:
		es.dbgmsg(0, message.strip())
	handleError(error_list)
sys.excepthook = except_hook

### register addon using the info dict like object and the email address ###
def registerAddon(info,email=None,language='en'):
	global gt
	es.dbgmsg(3,'Errorlib: Register addon %s' % info['basename'])
	reggedAddons[info['basename']] = {
		'email':email,
		'version':info['version'],
		'author':info['author'],
		'name':info['name'],
		'language':language,
		'errors':[]}
	if not gt:
		gt = True
		gamethread.delayed(3600,cleaner)

### unregisters the addon, forces an auto-prune and saves up some memory i guess ###
def unregisterAddon(basename):
	if type(basename) == dict:
		try:
			basename = basename['basename']
		except:
			pass
	try:
		autoPrune(basename)
		del reggedAddons[basename]
		es.dbgmsg(3,'Errorlib: Unregister addon %s' % basename)
	except:
		pass

### handles errors (scans them for addon name, checks them with regged addons, takes action according to cfg) ###
def handleError(error_list):
	es.dbgmsg(3,'Errorlib: Handle error')
	basenames = scanError(error_list)
	for basename in basenames:
		errorhash = md5gen(tracebackUnify(error_list,reggedAddons[basename]['version']))
		if not errorhash in reggedAddons[basename]['errors']:
			es.dbgmsg(3,'Errorlib: Error not sent yet')
			cfgMethod = configs('method')
			if cfgMethod != 'disabled':
				reggedAddons[basename]['errors'].append(errorhash)
				if cfgMethod == 'local':
					saveError(error_list,basename)
				elif cfgMethod == 'both' and reggedAddons[basename]['email']:
					sendError(error_list,basename).start()
					saveError(error_list,basename)
				elif reggedAddons[basename]['email']:
					sendError(error_list,basename).start()
				else:
					saveError(error_list,basename)

### reads the configs ###
def configs(var):
	cfg = {'method':'email','filetimeout':'5 d'}
	if os.path.exists("%s/cfg/errorlib.cfg" % gamedir):
		f = open("%s/cfg/errorlib.cfg" % gamedir, "r")
		for line in f:
			if (len(line) > 0) and (line[0:2] != '//') and (line != '\n'):
				line = line.replace("\n", "")
				line = line.replace("\t", " ")
				line = line.replace("  ", " ")
				linelist = line.split(" ", 1)
				cfg[linelist[0].replace("\"", "")] = linelist[1]
		f.close()
		value = cfg[var]
		return value
	return None

### scans a traceback string for registered addons involved in the error ###
def scanError(error_list):
	bns = []
	el = error_list
	el.reverse()
	for line in el:
		line = line.replace('\\','/')
		if ".py" in line:
			name = line[line.rfind('/') + 1:line.find('.py')]
			if reggedAddons.has_key(name) and name not in bns:
				bns.append(name)
	return bns

### save an error ###
def saveError(error,basename):
	es.dbgmsg(3,'Errorlib:Saving error')
	ts = time.strftime("%Y_%m_%d_%H_%M_%S")
	f = open("%s/addons/eventscripts/%s/%s.log" % (gamedir,basename,ts),'w')
	f.write('''Traceback:

%s

OS:

  %s
 
Addon Version:

  %s
 
Addon md5:

  %s
 
ES Version:

  %s
 
ES Noisy:

  %s
 
ES-Tools:

  %s
 
Loaded Addons:

  %s''' % (listToMultiline(error),
sys.platform,
reggedAddons[basename]['version'],
md5gen(basename),
str(es.ServerVar('eventscripts_ver')),
str(es.ServerVar('eventscripts_noisy')),
str(os.path.exists("%s/addons/es_tools.vdf" % gamedir)),
loadedAddonsList().replace(', ','\n')))
	f.close()
	
### send error via email ###
class sendError(Thread):
	def __init__(self,error,basename):
		Thread.__init__(self)
		es.dbgmsg(3,'Errorlib: init email sending')
		self.error = error
		self.email = reggedAddons[basename]['email']
		self.basename = basename
		self.language = reggedAddons[basename]['language']
	def run(self):
		es.dbgmsg(3,'Errorlib: running email sending')
		f = open('%s/addons/eventscripts/%s/%s.py' % (gamedir,self.basename,self.basename),'r')
		c = f.read()
		f.close()
		params = urllib.urlencode({
			'send':reggedAddons[self.basename]['email'],
			'author':reggedAddons[self.basename]['author'],
			'name':reggedAddons[self.basename]['name'],
			'traceback':listToMultiline(self.error),
			'traceback_antispam':tracebackUnify(self.error,reggedAddons[self.basename]['version']),
			'os':sys.platform,
			'version':reggedAddons[self.basename]['version'],
			'hash':md5gen(c),
			'esversion':str(es.ServerVar('eventscripts_ver')),
			'esnoisy':str(es.ServerVar('eventscripts_noisy')),
			'est':str(os.path.exists('%s/addons/es_tools.vdf' % gamedir)),
			'addons':loadedAddonsList()
		})
		es.dbgmsg(3,'Errorlib: open url')
		u = urllib.urlopen('http://www.ojii.ch/errorlib.php?basename=%s&lang=%s' % (self.basename,self.language),params)
		es.dbgmsg(3,'Errorlib: reading url')
		response = u.read()
		es.dbgmsg(3,'Errorlib: displaying response')
		es.dbgmsg(2,response)

### generates a md5 hash of a string ###
def md5gen(s):
	h = hashlib.md5()
	h.update(str(s))
	return h.hexdigest()

### converts the loaded addons module list into a string ###
def loadedAddonsList():
	s = ''
	for addon in es.addons.getAddonList():
		addon = str(addon).replace('\\','/')
		a = addon.find("'") + 1
		b = addon.find("'",a)
		c = addon.rfind('/') + 1
		d = addon.rfind('.py')
		s += '%s from %s, ' % (addon[a:b],addon[c:d])
	return s[:-2]
	
### converts a list into a multiline string (thanks SuperDave) ###
def listToMultiline(l):
	return '\n'.join(l)

### unifies the traceback (removes file paths,...) ###
def tracebackUnify(l,version):
	r = 'Version:%s;Type:%s;' % (version,l[0][:l[0].rfind(':')])
	for line in l:
		if 'File' in line:
			line = line.replace('\\','/')
			fc = line.find(',') + 1
			f = line[line.rfind('/') + 1:line.find('.py')]
			ln = line[line.find(' line ') + 6:line.find(',',fc)]
			fun = line[line.find(' in ') + 4:]
			r += 'File:%s,Line:%s,Function:%s;' % (f,ln,fun)
	r = r[:-1]
	if len(r) > 31:
		return md5gen(r)
	return r
	
### removes unloaded reggedAddons and autoprunes ###
def cleaner():
	es.dbgmsg(3,'Errorlib: Cleaning...')
	autoPrune()
	loaded = []
	for addon in es.addons.getAddonList():
		addon = str(addon)
		a = addon.find("'") + 1
		b = addon.find("'",a)
		nname = addon[a:b]
		if nname.count('.') != 0:
			names = addon[a:b].split('.')
		loaded.append(names[-1])
	toBeDeleted = []
	for addon in reggedAddons:
		if not addon in loaded:
			toBeDeleted.append(addon)
	for addon in toBeDeleted:
		es.dbgmsg(3,'Errorlib: Unregister addon %s' % addon)
		del reggedAddons[addon]
	gamethread.delayed(3600,cleaner)

### gets rid of old error logs ###
def autoPrune(basename=None):
	if basename:
		searchAndDelete("%s/addons/eventscripts/%s/" % (gamedir,basename))
	else:
		for addon in reggedAddons:
			searchAndDelete("%s/addons/eventscripts/%s/" % (gamedir,addon))

### Does the actual searching and deleting ###
def searchAndDelete(folder):
	t = time.time()
	for f in os.listdir(folder):
		if f.count('_') == 5 and f.endswith('.log'):
			if t > getTimeWithDiff(f[:-4]):
				os.remove("%s/%s" % (folder,f))

### converts the filename into time and adds the 10 days ###
def getTimeWithDiff(s):
	unitConvert = {'h':3600,'H':3600,'d':86400,'D':86400}
	a = time.strptime(s,"%Y_%m_%d_%H_%M_%S")
	b = time.mktime(a)
	cfgAmount,cfgUnit = configs('filetimeout').split(' ')
	c = float(cfgAmount) * unitConvert[cfgUnit]
	return b + c